from typing import List,Set,Tuple,Dict,Callable,Any

# External Modules
from copy   import deepcopy
from pprint import pformat,pprint
import networkx as nx                  # type: ignore
from MySQLdb import OperationalError # type: ignore

# Internal Modules
from dbgen.support.datatypes.table     import Table,Col,View
from dbgen.support.datatypes.func      import Func
from dbgen.support.datatypes.rule      import Rule
from dbgen.support.datatypes.misc      import onlyTest,xTest,fTest
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal,Text

from dbgen.support.utils               import (sqlexecute,sqlselect,ConnectInfo
                                                  ,mkInsCmd ,get_tab_id,topsort_with_dict)
from dbgen.support.create_meta         import create_empty
from dbgen.core.lists                  import flatten
#########################################################################################

class DBG(object):
    """
    Container for the three necessary components to define a database update
    """
    def __init__(self
                #,funcs      : List[Func]
                ,tables     : List[Table]
                ,rules      : List[Rule]
                ,views      : List[View]  = []
                ) -> None:
        # Functions
        #self.funcs     = funcs

        # Schema
        self.views      = views
        self.tables     = self.order_tables(tables)

        self.fkdict = {} # type: Dict[str,Tuple[str,str]]
        for t in self.tables:
            for fk in t.fks:
                for c1,c2 in zip(fk.column,fk.target):
                    if '%s.%s'%(t.name,c1) not in self.fkdict: #HACK - how to handle when a column is part of multiple FKs?
                        self.fkdict['%s.%s'%(t.name,c1)] = (fk.table,'%s.%s'%(fk.table,c2))

        # Rules
        self.rules = rules # unordered
        self.rules = self.order_rules(rules)

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def str_to_tab(self,tabname:str)->Table:
        for t in self.tables:
            if t.name == tabname: return t
        raise ValueError('cannot find table '+tabname)

    def view_deps(self
                 ,tablist:Set[str]
                 ,collist:Set[str]
                 ) -> Tuple[Set[str],Set[str]]:
        """
        REMOVE views from the parsed tables that are dependencies.  These views
        depend on tables and columns themselves, so return
        """
        dep_dict =  {v.name:v.deps() for v in self.views}

        tabs,cols   = set(),set() # type: Set[str],Set[str]

        for tab in tablist:
            if tab in dep_dict.keys():
                ts,cs = dep_dict[tab]
                tabs.update(ts)
                cols.update(cs)
            else:
                tabs.add(tab)

        return tabs,cols | collist

    def derive_deps(self,t : Rule)->List[str]:
        """
        For a given rule, first determine everything it depends on (parent tabs)
        Then iterate through every other rule and determine what it affects (child tabs)
        If there is any intersection, then this rule depends on that one

        This has to be done here (rather than at the Rule level) because
        there is dependency on dbg.views and dbg.tables (FKs)
        """
        debugFrom = 'cell_info'
        debugTo   = None # 'populate_struct_traj_atom_cell'

        output, n = [],t.name

        raw_parent_tabs,raw_parent_cols,child_tabs,child_cols = t.deps
        viewed_p_tabs,viewed_p_cols   = self.view_deps(raw_parent_tabs,raw_parent_cols) # correct for effect of Views
        viewed_c_tabs,viewed_c_cols = self.view_deps(child_tabs,child_cols)           # correct for effect of Views

        fk_tabs,fk_cols = self.get_fk_deps(child_cols)

        produced_tab_pks = set(flatten([['%s.%s'%(tab,c.name)
                                 for c in self.str_to_tab(tab).cols if c.pk]
                                    for tab in viewed_c_tabs]))

        # Finally, determine all tables/cols that this Rule depends on
        parent_tabs =  (viewed_p_tabs | fk_tabs) - child_tabs
        parent_cols = ((viewed_p_cols | fk_cols) - viewed_c_cols) - produced_tab_pks

        for other in self.rules:
            o = other.name
            if o != n:
                _,_,other_child_tabs,other_child_cols = other.deps
                disjoint = ( other_child_tabs.isdisjoint(parent_tabs)
                           & other_child_cols.isdisjoint(parent_cols))
                if debugFrom==n and debugTo==o:
                    print("DEBUGGING DEPENDENCIES\n#######################\n")
                    if t.query:
                        print("%s's query \n"%n+pformat(t.query))
                    else:
                        print("%s's metatemp "%n)
                        for b in t.metatemp.blocks: print(b.text)
                    print("%s's metatemp "%o)
                    for b in other.metatemp.blocks: print(b.text)
                    print('\nRESULTS\n########\n')
                    print("%s's parent_tabs \n\t"%n   + pformat(parent_tabs))
                    print("%s's parent_cols \n\t"%n   + pformat(parent_cols))
                    print("%s's child_tabs\n\t"%o + pformat(other_child_tabs))
                    print("%s's child_cols\n\t"%o + pformat(other_child_cols))
                    import pdb;pdb.set_trace()
                if not disjoint:
                    output.append(other.name)

        return output

    def get_fk_deps(self
                   ,colstrs : Set[str]
                   ) -> Tuple[Set[str],Set[str]]:
        """
        If we insert into TAB.COL and that is a FK to TAB2.COL2, then we have
        another dependency we need to account for
        """
        fktab,fkcol = set(),set()
        deps = map(self.fkdict.get,colstrs)
        for d in deps:
            if d is not None:
                fktab.add(d[0])
                fkcol.add(d[1])
        return (fktab,fkcol)

    @staticmethod
    def order_tables(tabs : List[Table])->List[Table]:
        """
        Order the tables, so that they can be created (in MySQL) without
        violating FK constraints
        """
        G = nx.DiGraph()
        G.add_nodes_from([t.name for t in tabs])
        tab_dict = {t.name:t for t in tabs}
        for t in tabs:
            for f in t.fks:
                G.add_edge(f.table,t.name)
        return topsort_with_dict(G,tab_dict)

    def order_rules(self,ts : List[Rule]) -> List[Rule]:
        """
        Sort rules such that dependencies in queries are respected.
        """
        G = nx.DiGraph()
        G.add_nodes_from([t.name for t in ts])
        t_dict = {t.name:t for t in ts}
        for t1 in ts:
            deps = self.derive_deps(t1)
            for t2 in ts:
                if t2.name in deps:
                    G.add_edge(t2.name,t1.name)
        return topsort_with_dict(G,t_dict)


    def make_metatables(self,conn : ConnectInfo)->None:
        """
        Initialize metatables
        """
        metatables = [Table('META_rule'
                        ,cols = [Col('id',pk=True,auto=True)
                                ,Col('method',Varchar(),nn=True)
                                ,Col('status',Varchar(),nn=True)
                                ,Col('runtime',Decimal())
                                ,Col('n_inputs')
                                ,Col('rate',Decimal())
                                ,Col('error',Text())])
                      ,Table('META_dependencies'
                          ,cols = [Col('id',pk=True,auto=True)
                                ,Col('parent',Varchar(),nn=True)
                                ,Col('child',Varchar(),nn=True)])]

        for ta in metatables:
            sqlexecute(conn,'DROP TABLE IF EXISTS '+ta.name)
            sqlexecute(conn,ta.create_table())

        q  = mkInsCmd('META_rule',['method','status'],sqlite=False)
        dq = mkInsCmd('META_dependencies',['child','parent'],sqlite=False)
        for tr in self.rules:
            sqlexecute(conn,q,[tr.name,'initialized'])
            for dep in self.derive_deps(tr):
                sqlexecute(conn,dq,[tr.name,dep])

    def run_all(self
               ,conn    : ConnectInfo
               ,only    : Set[str]  = set()
               ,xclude  : Set[str]  = set()
               ,reset   : bool = False
               ,add     : bool = False
               ,verbose : bool = True
               ,parallel: bool = True
               ) -> None:
        """
        Execute all rules
        """
        print('starting dbgen...')

        self.make_metatables(conn)

        if reset:
            for ta in reversed(self.tables):
                try:
                    sqlexecute(conn,'DROP TABLE IF EXISTS '+ta.name)
                except:
                    raise ValueError("""\t\t\tIn mysql workbench do the following:
                                set session foreign_key_checks = 0 ;
                                <DROP ALL TABLES USING THE SIDEBAR>
                                set session foreign_key_checks = 1 ;""")
        elif add:
            print('trying to add new columns...')
            for ta in self.tables:
                for sqlexpr in ta.add_cols():
                    try:
                        sqlexecute(conn,sqlexpr)
                    except OperationalError:
                        pass
            print('\tdone trying to add new columns.')

        for ta in self.tables:
            sqlexecute(conn,ta.create_table(if_not=True))
            for chk,stmt in ta.add_indices(): #get a SQL error when I try this...
                if not sqlselect(conn,chk):
                    sqlexecute(conn,stmt)

        for v in self.views:
            sqlexecute(conn,v.create_view())

        for rule in self.rules:
            testdict = {onlyTest : [only]
                       ,xTest    : [xclude]
                       ,fTest    : [conn]}
            run = True # flag for passing all tests
            for test,args in testdict.items():
                test_output = test.run_test(rule,*args) # type: ignore
                if test_output is not True:
                    print('\tNot running rule ',rule.name)
                    rule.update_status(conn,test_output)
                    run = False
                    break
            if run:
                print('\tRunning rule ',rule.name)
                rule.update_status(conn,'running')
                rule.run_rule(conn
                           ,verbose
                           ,parallel)

    def create_dbg(self,sqlite_pth:str)->None:
        """
        Creates a whole database using a Python DBG instance
        """
        create_empty(sqlite_pth,local=False)
        for t in self.tables:
            t.add_table(sqlite_pth)
        for t in self.tables:
            tab_id = get_tab_id(sqlite_pth,t.name)
            t.add_fks(sqlite_pth,tab_id)
        #for func in self.funcs:
        #    func.add_func(sqlite_pth)
        for tr in self.rules:
            tr.add_rule(sqlite_pth)
        for v in self.views:
            v.add_view(sqlite_pth)

    def test_input(self
                  ,tr_name : str
                  ,inp     : dict
                  ) -> None:
        t = [tr for tr in self.rules if tr.name == tr_name][0]
        print(t.template.apply_template(inp))

    def test_meta(self)->None:
        """
        Create meta.db and confirm that conversion process is reversible
        """
        from dbgen.support.extract_python import extract_dbg

        self.create_dbg('meta.db')
        dbg_ = extract_dbg('meta.db')
        if self!=dbg_:
            self.diff(dbg_) # print out differences, if any
        assert self == dbg_, 'Something wrong with input or DBG <-> meta.db'

    def diff(self,dbg:'DBG')->None:
        """
        If one expects to DBG's to be equal and they aren't, diff them.
        """
        attrs = ['funcs','tables','views','rules']
        fNames,tNames,vNames,trNames = [[[x.name for x in getattr(d,attr)] for d in [self,dbg]] for attr in attrs]
        # if len(self.funcs)!=len(dbg.funcs):
        #     print("Different number of funcs: \n{} \n{} ".format(*fNames))
        #     return None
        # else:
        #     for f,f2 in zip(self.funcs,dbg.funcs):
        #         assert f.name==f2.name, "Different function orderings {}\n{}".format(*fNames)
        #         if f!=f2:
        #             print('Functions differ for %s\n%s\n%s'%(f.name,pformat(f),pformat(f2)))
        #             return None
        if len(self.tables)!=len(dbg.tables):
            print("Different number of tables: \n{}\n{}".format(*tNames))
            return None
        else:
            for t,t2 in zip(self.tables,dbg.tables):
                assert t.name==t2.name, "Different table ordering {} vs \n {}".format(*tNames)
                if t!=t2:
                    print('Tables differ for '+t.name)
                    if t.name!=t2.name:
                        print('\tNames differ \n\t %s \n%s '%(t.name,t2.name))
                    elif t.desc!=t2.desc:
                        print('\tDescriptions differ\n\t %s \n%s '%(t.desc,t2.desc))
                    for c,c2 in zip(t.cols,t2.cols):
                        if c!=c2:
                            print('\tColumns differ \n\t %s \n%s '%(c,c2))
                    for fk,fk2 in zip(t.fks,t2.fks):
                        if fk!=fk2:
                            print('\FKs differ \n\t %s \n%s '%(fk,fk2))
                    return None
        if len(self.views)!=len(dbg.views):
            print("Different number of views: \n{}\n{}".format(*vNames))
            return None
        else:
            for v,v2 in zip(self.views,dbg.views):
                assert v.name==v2.name, "Different ordering of views  \n{}\n{}".format(*vNames)
                if v!=v2:
                    print("Views differ for %s \n%s \n%s"%(v.name,pformat(v),pformat(v2)))
        if len(self.rules)!=len(dbg.rules):
            print("Different number of rules: {}\n{} ".format(*trNames))
            return None
        else:
            for tr,tr2 in zip(self.rules,dbg.rules):
                assert tr.name==tr2.name, "ordering different {} \n {}".format(*trNames)
                if tr!=tr2:
                    print('Rules differ for %s \n%s\%s'%(tr.name,pformat(tr),pformat(tr2)))
                    return None
        print("Tables don't differ")



from typing import List,Tuple
from os import listdir,environ,walk
from fnmatch import fnmatch
from os.path import join,exists
from glob import glob
from subprocess import check_output,getstatusoutput

def get_qe_logfiles(rootpath:str) -> Tuple[List[str],List[str],List[str]]:
    """
    Searches for DFT calculation folders. Returns unzipped (stordir,logfile,code) triples

    Requirements to be flagged as a DFT job:
        - There must exist at least one .traj file in the directory
        - If there's also a 'runtime.json' file, we'll assume it's a CataLog job
        - If there's a DDEC analysis file, we'll assume it's a Chargemol job
        - If there's a subdirectory with one 'pw.inp' and 'log' file, it's a QE job
        - If there exists a file in the directory with a particular signature,
            we identify it as a GPAW job.
    """
    ############################
    # Initialize Variables
    #---------------------
    output = []
    user   = environ['SHERLOCK_USERNAME']

    # Main Program
    #-------------

# dirname dirname dirname dirname
    cmd = r"find {0} -name pw.inp -printf '%h\n'".format(rootpath)
    for calcdir in check_output(cmd).split():
        slash = calcdir.rfind('/')
        stordir = calcdir[:slash]
        output.append((stordir,calcdir,'quantumespresso'))

    if output:
        return tuple(map(list,zip(*output))) # type: ignore
    else:
        print('warning, no log files found in %s'%rootpath)
        return [],[],[]



from typing import Dict,Optional,Tuple
import json
from dbgen.core.parsing import parse_line

def get_nbands_gpaw(files_params  : Tuple[str,str]
              ) -> Optional[int]:
    """
    If we do have a params json, try to get nbands from it.
    """
    params_raw = files_params[1]
    if not params_raw:
        return None
    else:
        params = json.loads(params_raw)
        return int(params.get('nbands'))     # no conceivable way of getting this from log



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_dipol_qe(log:str) -> Optional[str]:
    """
    Location of dipole initial guess ... don't know how to parse (if even possible)
    """
    return None



# External Modules
from typing import Optional,List,Union,Any,Tuple,Set
from pprint import pformat
# Internal Modules
from dbgen.support.sqlparsing         import parse_dependencies
from dbgen.support.utils              import sqlite_select,sqlite_execute,mkInsCmd,namespaceCmd,get_tab_id
from dbgen.support.datatypes.sqltypes import SQLType,Int

"""
Defines Table, Col, FK, and View

"""

################################################################################
getColCmd  = """SELECT col.col_id FROM tab
                JOIN col USING (tab_id)
                WHERE tab.tab_id = ? AND col.name = ? """
################################################################################

class Table(object):
    """
    - Tables are used to collect information about a class of entities/objects
    """
    def __init__(self
                ,name  : str
                ,desc  : Optional[str]  = 'No description'
                ,cols  : List['Col']     = []
                ,fks   : List['FK']     = []
                ) -> None:
        assert any([x.pk for x in cols]), "need at least one primary key column for "+name
        self.name = name
        self.cols = self.mkFKcols(cols,fks)
        self.desc = desc
        self.fks  = fks

    def src(self)->str:
        cols = '\n\t\t,'.join([c.src() for c in self.cols])
        fks = '\n\t,fks = [%s]'%'\n\t\t,'.join([fk.src() for fk in self.fks]) if self.fks else ''
        return "Table('{0}'\n\t,desc=\"\"\"{1}\"\"\"\n\t,cols=[{2}]{3})".format(self.name,self.desc,cols,fks)

    def __str__(self)->str:
        return pformat(self.__dict__)
    def __repr__(self)->str:
        return str(self)
    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    @staticmethod
    def mkFKcols(cols : List['Col']
                ,fks  : List['FK']
                ) -> List['Col']:
        """
        Let's be honest: we're lazy!
        We don't want to have to rewrite a vanilla Foreign Key that's already
        been specified elsewhere. Automatically add it to the column list if we
        don't see it there.
        """
        names = [c.name for c in cols]
        for fk in fks:
            for colname in fk.column:
                if colname not in names:
                    cols.append(Col(colname,Int(),False,False,False))
        return cols


    def add_cols(self)->List[str]:
        return ["ALTER TABLE {0} ADD {1}".format(self.name,c.create_col(sqlite=False))
                for c in self.cols if not c.pk]


    def create_table(self
                    ,sqlite  : bool  = False
                    ,if_not  : bool  = False
                    ) -> str:
        """
        Returns the statement necessary to create table, MySQL or SQLite
        """
        cols           = [c.create_col(sqlite) for c in self.cols]
        constraints   = ( [self.pkConstraint()]
                        + self.uniqConstraint()
                        + [str(f) for f in self.fks])

        if_not_str    = 'IF NOT EXISTS' if if_not else ''
        mbConstraints = '' if len(constraints)==0 else '\n\t,'.join(constraints)

        cmd = "CREATE TABLE %s %s\n\t(%s%s)"%(if_not_str
                                             ,self.name
                                             ,'\n\t,'.join(cols)
                                             ,mbConstraints)
        return cmd

    def pkConstraint(self)->str:
        pkCols = [c.name for c in self.cols if c.pk]
        return "\n\t,PRIMARY KEY (%s)"%(','.join(pkCols))

    def uniqConstraint(self)->List[str]:
        ucs = [c.name for c in self.cols if c.uniq]
        if ucs == []:
            return []
        else:
            return ["UNIQUE (%s)"%','.join(ucs)]

    def add_table(self,sqlite_pth:str)->None:
        """
        Adds a table, then its columns
        """
        name = self.name
        cmd  = mkInsCmd('tab',['name','desc'],sqlite=True)
        sqlite_execute(sqlite_pth,cmd,[name,self.desc])
        sqlite_execute(sqlite_pth,namespaceCmd,[name,'Table'])

        tab_id = sqlite_select(sqlite_pth,'SELECT tab_id FROM tab WHERE name=?',[name])[0][0]

        ins_cols = ['col_id','tab_id','name','type','nnull','pk','uniq'
                   ,'auto','ind','virt','dfault']
        for col_id,c in enumerate(self.cols):
            binds    = [col_id,tab_id,c.name,str(c.type),c.nn,c.pk
                       ,c.uniq,c.auto,c.ind,c.virt,str(c.default)]

            cmd = mkInsCmd('col',ins_cols,sqlite=True)
            sqlite_execute(sqlite_pth,cmd,binds)
            sqlite_execute(sqlite_pth,namespaceCmd,['%s.%s'%(name,c.name),'Column'])

    def add_fks(self
               ,sqlite_pth : str
               ,tab_id     : int
               ) -> None:
        """
        Add foreign keys to meta.db (ALL tables need to be added, first)
        """
        for i,fk in enumerate(self.fks):
            insQ = mkInsCmd('fk',['tab_id','fk_id'],sqlite=True)
            sqlite_execute(sqlite_pth,insQ,[tab_id,i])
            fk.add_fk(sqlite_pth,tab_id,i)

    def get_fk(self, t2: 'Table')->'FK':
        for fk in self.fks:
            if fk.table == t2.name:
                return fk
        raise ValueError('Could not find FK between %s and %s'%(self,t2))

    def add_indices(self)->List[Tuple[str,str]]:
        return [("""SELECT COUNT(1) IndexIsThere FROM INFORMATION_SCHEMA.STATISTICS
                    WHERE table_schema=DATABASE() AND table_name='%s' AND index_name='%s';"""%(self.name,c.name)
                ,'ALTER TABLE %s ADD INDEX (%s)'%(self.name,c.name))
                    for c in self.cols if c.ind]


class Col(object):
    """
    typ = SQL datatype {int,str,float}
    nn  = NOT NULL constraint
    pk  = PRIMARY KEY constraint
    uniq = include column in a table-wide UNIQUE constraint
    Notes:
    - SQLite doesn't automatically enforce a NOT NULL constraint on PKs
        ...but we want that.
    """
    def __init__(self
                ,name : str
                ,typ  : SQLType = Int() #str  = 'int'
                ,nn   : bool = False
                ,pk   : bool = False
                ,uniq : bool = False
                ,auto : bool = False
                ,ind  : bool = False
                ,virt : Optional[str] = None
                ,default : Optional[Any] = None
                ) -> None:
        self.name = name
        self.type = typ
        self.nn      = nn or pk
        self.pk      = pk
        self.uniq = uniq
        self.auto = auto
        self.ind  = ind
        self.virt = virt
        self.default = default

    def src(self)->str:
        typ = '' if self.type == 'int' else ",typ='%s'"%self.type
        nn  = '' if self.pk or not self.nn else ',nn=True'
        pk  = '' if not self.pk else ',pk=True'
        uniq= '' if not self.uniq else ',uniq=True'
        auto= '' if not self.auto else ',auto=True'
        ind = '' if not self.ind  else ',ind=True'
        virt = '' if not self.virt else ',virt=%s'%self.virt
        default = '' if self.default is None else ',default=%s'%repr(self.default)
        args = [typ,nn,pk,uniq,auto,ind,virt,default]
        return "Col('%s'%s)"%(self.name,''.join(args))

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def create_col(self
                  ,sqlite : bool = False
                  )->str:
        """
        Create statement for column when creating a table.
        NO idea why I need to include 'null' there for the sqlite typechecking...
        """

        dt = str(self.type) # self.typToSQLType(self.type).lower()
        nn = 'NOT NULL' if self.nn else ''
        auto = 'AUTO_INCREMENT' if self.auto else ''
        virt = '' if self.virt is None else "AS (%s)"%self.virt
        dflt = '' if self.default is None else "DEFAULT %s"%(self.default)
        # if sqlite:
        #     if 'var' in dt:
        #         check_dt = 'text' # HACK
        #     else:
        #         check_dt = dt
        #
        #     return "{0} \t{1} {2} {3} CHECK (typeof({0}) in ('{4}','null'))".format(self.name,dt,nn,auto,check_dt)
        # else:
        return "{0} \t{1} {2} {3} {4} {5}".format(self.name,dt,nn,auto,virt,dflt)


class FK(object):
    """
    Representation of a foreign key relationship.
    Use lists for composite foreign keys.

    FOREIGN KEY (job_id) REFERENCES traj (job_id)
    """
    def __init__(self
                ,fkc : Union[List[str],str]
                ,fkt : str
                ,tar : Optional[Union[List[str],str]] = None
                ) -> None:
        self.column  = fkc if isinstance(fkc,list) else [fkc]
        self.table   = fkt
        if tar is None:
            self.target = self.column
        else:
            self.target  = tar if isinstance(tar,list) else [tar]

        assert len(self.column) == len(self.target)

    def __str__(self) -> str:
        col,tab,tar = ','.join(self.column),self.table,','.join(self.target)
        return "FOREIGN KEY (%s) REFERENCES %s (%s)"%(col,tab,tar)

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self,other : object) -> bool:
        return self.__dict__ == other.__dict__


    def add_fk(self
              ,sqlite_pth : str
              ,tab_id     : int
              ,fk_id        : int
              ) -> None:
        """
        Adds FK relation to SQLite meta.db
        """
        for fromC,toC in zip(self.column,self.target):
            from_col = sqlite_select(sqlite_pth,getColCmd,[tab_id,fromC])[0][0]
            try:
                to_tab_id = get_tab_id(sqlite_pth,self.table)
                to_col = sqlite_select(sqlite_pth,getColCmd,[to_tab_id,toC])[0][0]
            except IndexError:
                raise ValueError("Could not find table %s with col %s"%(self.table,toC))

            cmd = mkInsCmd('fk_cols',['tab_id','fk_id','col_id','to_tab','to_col'],sqlite=True)
            sqlite_execute(sqlite_pth,cmd,[tab_id,fk_id,from_col,to_tab_id,to_col])


    def src(self)->str:
        targ = '' if self.target == self.column else ',%s'%self.target
        return "FK({0},'{1}'{2})".format(self.column,self.table,targ)

class View(object):
    def __init__(self,name:str,query:str)->None:
        self.name  = name
        self.query = query

    def src(self)->str:
        return "View('%s','%s')"%(self.name,self.query)
    def __str__(self)->str:
        return pformat(self.__dict__)
    def __repr__(self)->str:
        return str(self)
    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__
    def create_view(self)->str:
        return "CREATE OR REPLACE VIEW %s AS %s"%(self.name,self.query)
    def deps(self)->Tuple[Set[str],Set[str]]:
        return parse_dependencies([self.query])[:2]
    def add_view(self,sqlite_pth:str)->None:
        cmd = mkInsCmd('views',['name','query'],sqlite=True)
        sqlite_execute(sqlite_pth,cmd,[self.name,self.query])
        sqlite_execute(sqlite_pth,namespaceCmd,[self.name,'View'])



from typing                             import Tuple
from pymatgen.core.surface              import generate_all_slabs, SlabGenerator        # type: ignore
from pymatgen.symmetry.analyzer         import SpacegroupAnalyzer                       # type: ignore
from pymatgen                           import Lattice, Structure                       # type: ignore
from pymatgen.core.surface              import get_recp_symmetry_operation
from pymatgen.analysis.elasticity       import get_uvec                                 # type: ignore
from monty.fractions                    import gcd_float                                # type: ignore
from ase.io                             import read                                     # type: ignore
from pymatgen.io.ase                    import AseAtomsAdaptor                          # type: ignore
from ase                                import Atoms                                    # type: ignore
from ase.constraints import FixAtoms                                                    # type: ignore
import numpy as np                                                                      # type: ignore
from dbgen.core.lists import normalize_list

def surface_bulkstruct(atoms      : Atoms
                      ,resolution : float = 0.05
                      ,top_pad    : float = 5.0
                      ,symprec    : float = 0.2
                      ) -> Structure :
    """
    Drop the c lattice vector and look for highest symmetry structure, which is
    the representative of the bulk structure
    """
    print('starting surface bulkstruct with atoms = '+str(atoms))
    # helper class to break out of nested For loops
    #----------------------------------------------
    class BreakIt(Exception): pass

    # Identify constrained atoms
    #---------------------------
    fixed_inds = []
    if atoms.constraints:
        for constraint in atoms.constraints:
            if isinstance(constraint,FixAtoms):
                fixed_inds.extend(list(constraint.get_indices()))
    if fixed_inds==[]:
        fixed_inds = list(range(len(atoms)))

    # Only return structures with same composition as bulk
    #------------------------------------------------------
    comp_norm = sorted(normalize_list([x.number for x in atoms if x.index in fixed_inds]))

    # Initialize other variables
    #---------------------------
    highest_symmops,best_sgn,best_structure = 0,0,None
    slab       = AseAtomsAdaptor().get_structure(atoms)
    index_list = range(len(atoms))
    c_uvec     = get_uvec(slab.lattice.matrix[2])

    # Get unique c values to translate by
    #------------------------------------
    c_values = slab.frac_coords[:, 2].round(4)
    c_values = np.unique(c_values)
    c_values = sorted(c_values)
    c_values.pop(-1) # Probably don't need the top one

    for c_value in c_values: # for each atom

        # Translate sites so that this site is at the bottom of the cell
        #---------------------------------------------------------------
        test_struct = slab.copy()
        test_struct.translate_sites(index_list,[0,0,-c_value],to_unit_cell=False)

        # Remove all sites out of lower bound
        #------------------------------------
        above = lambda s: s.frac_coords[2] >= -0.0001
        test_struct = Structure.from_sites(list(filter(above,test_struct))) #[s for s in test_struct if s.frac_coords[2]>=-0.0001])

        # Resize lattice
        #---------------
        slab_c_mag = slab.lattice.c
        max_c_mag = slab_c_mag * np.max(test_struct.frac_coords[:, 2]) + top_pad # scale by max c coordinate
        c_mags = np.arange(2.0, max_c_mag, resolution).tolist()

        for c_mag in reversed(c_mags):
            new_matrix    = slab.lattice.matrix
            new_matrix[2] = c_uvec * c_mag
            new_lattice   = Lattice(new_matrix)

            # Cut off top
            #------------
            new_sites         = [s for s in test_struct if new_lattice.get_fractional_coords(s.coords)[2] < 1]
            inner_test_struct = Structure(new_lattice, [s.species_string for s in new_sites],
                                          [s.coords for s in new_sites], coords_are_cartesian=True)

            new_comp_norm = sorted(normalize_list(inner_test_struct.atomic_numbers))
            if new_comp_norm == comp_norm:
                try:
                    # Search for overlapping atoms
                    for i,a in enumerate(inner_test_struct):
                        if a.z < 0.3:
                            for j,b in enumerate(inner_test_struct):
                                if i!=j and a.distance(b)<0.3:
                                    raise BreakIt

                        # Test if new structure is the 'best' one seen so far
                        try:
                            sga = SpacegroupAnalyzer(inner_test_struct, symprec=symprec)
                            num_symmops = len(sga.get_symmetry_operations())
                            if len(sga.get_symmetry_operations()) > highest_symmops:
                                best_structure = inner_test_struct
                                highest_symmops = num_symmops
                                best_sgn        = sga.get_space_group_number()
                        except TypeError:
                            pass
                except BreakIt:
                    pass
    return best_structure

if __name__=='__main__':
    import sys
    a = read(sys.argv[1])
    print(surface_bulkstruct(a))



from typing import Tuple
from dbgen.core.parsing import parse_line

def get_maxstep_qe(trip:Tuple[str,str,str]) -> int :
    """
    We need the dict instead of logfile because this info is only in pw.inp?
    """
    _,pwinp,_  = trip
    parsed = parse_line(pwinp,'electron_maxstep',0)
    if parsed is None:
        raise ValueError('malformed pw.inp?')
    else:
        return int(parsed.split('=')[1][:-1])



from typing import Tuple,Optional

def get_gamma_vasp(trip:Tuple[str,str,str]) -> Optional[int]:
    """
    Checks for gamma in KPOINTS
    """
    _,_,kptcar = trip
    return int('gamma' in kptcar)



from typing import List,Tuple
import json
import os

################################################################################

def parse_mendeleev()->Tuple[List[int],List[int],List[str]
                                ,List[float],List[int],List[str]]:
    """
    Extracts information of elements
    """
    rootpath = os.environ['DBGEN_ROOT']

    with open(os.path.join(rootpath,'data/element.json'),'r') as f:
        elems = json.load(f)
    ############################
    cols = ['atomic_number','atomic_weight','spacegroup','pointgroup'] # etc

    allcols = ['atomic_number', 'symbol', 'name', 'atomic_weight','atomic_radius'
              , 'phase','evaporation_heat', 'pointgroup','spacegroup'
              , 'melting_point', 'metallic_radius', 'vdw_radius'
              , 'density', 'en_allen' , 'is_radioactive'
              , 'lattice_structure' , 'fusion_heat'
              , 'econf', 'period', 'covalent_radius_bragg'
              , 'geochemical_class', 'abundance_crust', 'heat_of_formation'
              , 'electron_affinity', 'atomic_volume',  'boiling_point'
              , 'proton_affinity', 'covalent_radius_slater'
              , 'lattice_constant', 'dipole_polarizability'
              , 'en_ghosh', 'thermal_conductivity', 'group_id', 'en_pauling'
              , 'gas_basicity'
              ,'abundance_sea']

    output = []
    for e in elems:
        output.append([e.get(k) for k in allcols])

    return tuple(map(list,zip(*output))) # type: ignore



from typing import Dict

from dbgen.core.parsing import parse_line

def parse_xc_vasp(outcar:str) -> str:
    """
    Finds the exchange correlation functional
    """
    parsed = parse_line(outcar,'GGA  ',0)
    if parsed is None:
        raise ValueError('malformed OUTCAR? no "GGA  " string found')
    else:
        gga = parsed.split()[2] # <name> <equals sign> <GGA name>
        if gga == 'PE':
            return 'PBE' # need check for if BEEF?
        else:
            raise NotImplementedError('New functional in vasp? '+parsed)



from networkx import node_link_data,MultiGraph # type: ignore
from json import dumps                         # type: ignore

def graph_to_json(G : MultiGraph)->str:
    """
    Serialize a MultiGraph
    """
    return dumps(node_link_data(G))



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_spinpol_gpaw(log:str) -> Optional[int]:
    """
    Docstring
    """

    parsed = parse_line(log,'spinpol',0)
    if parsed is None:
        return 0
    else:
        return int('True' in parsed)



#External
from typing import Dict,Any,Tuple,Union,Set,List
# Internal
from dbgen.support.datatypes.sqltypes   import SQLType,Int,Varchar,Text,Decimal
from dbgen.support.datatypes.attr         import Attr



#############################################################################
# Constants
class NotNull(object): pass
NOTNULL = NotNull()

class DEFAULT(object):
    def __init__(self,val:Any)->None:
        self.val = val
################################################################################
################################################################################
################################################################################

class Object(object):
    """
    Superclass for user-defined objects (which correspond to DB tables)

    _parents
        -- SUFFICIENT and NECCESARY objects needed to specify an instance
          (e.g. for 'Atom', 'Struct' and 'Element' (+ 'atom_id')
           are both necessary, but Struct+atom_id are sufficient.
    _components
        --
    <not starting with '_'>
        -- normal attribute
    """
    universe = {}    # type: Dict[str,Object]
    todo     = set() # type: Set[str]

    def __init_subclass__(cls      : 'Object'  # type: ignore
                         ,**kwargs : dict
                         ) -> None:
        """
        Upgrade a user-defined class given the state of the universe
        """
        if kwargs:
            raise ValueError('kwargs? '+str(kwargs))
        else:
            self._init_subclass(cls)

    @staticmethod
    def _init_subclass(cls:'Object')->None:
        uni  = cls.universe          # type: ignore
        todo = cls.todo              # type: ignore
        name = cls.__name__.lower()  # type: ignore
        if name in uni:
            print('Warning: this class was already defined for this universe')
            return None # No repeats allowed
        else:
            cd              = cls.__dict__
            parents         = cd.get('_parents',[])
            cls._parents    = parents
            components      = cd.get('_components',[])
            cls._components = components
            many_to_one     = cd.get('_many_to_one',False)
            cls._many_to_one= many_to_one
            cls._attrs      = [cls.mkAttr(k,v) for k,v in cd.items() if k[0]!='_']
            for a in cls._attrs:
                setattr(cls,a.name,a)

            cls._pks = [] # type: List[Attr]

            if ((not parents) or many_to_one):
                cls._pks.append(Attr(name+'_id',pk=True,auto=True))


            for p in parents:
                assert isinstance(p,(Object,str))
                if isinstance(p,Object) or p.lower() in uni:
                    if isinstance(p,Object): parent = p
                    else: parent = uni[p.lower()]

                    cls._process_parent(parent)


                else:
                    todo.add(p.lower())
            #import pdb;pdb.set_trace()
            uni[name] = cls
            if name in todo:
                todo.remove(name)

    def _process_parent(self,par:'Object')->None:
        print('processing parent of '+str(self))
        print('par is '+str(par))
        pass

    def __init__(self,alias:str)->None:
        self.alias=alias

    @staticmethod
    def mkAttr(name:str,data:Union[SQLType,Tuple])->Attr:
        if isinstance(data,tuple):
            types = (SQLType,NotNull,DEFAULT)
            assert all([isinstance(x,types) for x in data])
            dtype = [d for d in data if isinstance(d,SQLType)][0]
            return Attr(name,dtype=dtype,nnull=NOTNULL in data)
        else:
            return Attr(name,dtype=data)
################################################################################
################################################################################
################################################################################
class Structure(Object):
    """
    Chemical structure defined in periodic cell
    """
    raw     = Text(),NOTNULL

class Element(Object):
    """
    Chemical element
    """
    mass    = Decimal(),NOTNULL

class Atom(Object):
    """
    An atom in a specific chemical structure
    """
    _parents    = [Structure]
    _components = [Element]
    x           =  Decimal(),NOTNULL
    y           =  Decimal(),NOTNULL
    z           =  Decimal(),NOTNULL



a1 = Atom('1')
a2 = Atom('2')

import pdb;pdb.set_trace()



from typing import Tuple
from json import loads
def get_catalog_facet(paramdict:str)->Tuple[int,int,int]:
    """
    Extract facet from params.json
    """
    params = loads(paramdict)
    facet = params.get('facet')
    if facet is None:
        facet = params.get('facet_json')
        if facet is None:
            raise ValueError("Cannot find facet in params.json")
    return tuple(loads(facet))



from typing import Union,List,Dict,Any,Callable,Tuple,Optional

# External Modules
import traceback
import networkx as nx         # type: ignore
from copy   import deepcopy
from pprint import pformat
# Iternal Modules
from dbgen.support.datatypes.func     import Func
from dbgen.support.datatypes.misc     import ExternalError,Arg
from dbgen.support.utils              import (mkInsCmd,sqlite_execute
                                                    ,func_to_ind,sqlite_select
                                                    ,topsort_with_dict)

##############################################################################
class Temp(object):
    """
    - A Template (Temp) object specifies a composition of functions which
        expand the namespace generated by a query.

    - These are related by a set of function *instances* (a function can be
        instantiated multiple times with different arguments).

    - For every function instance, we need to specify where it gets its inputs
        by referring to the namespace at that moment in time.
    """

    def __init__(self
                ,tempfuncs : List['TempFunc']   = []
                ,constants : Dict[str,'Const'] = {}
                ) -> None:
        self.tempfuncs = self.order_tempfuncs(tempfuncs)
        self.constants = constants

    def src(self)->str:
        tfs   = '\n\t\t,'.join([tf.src() for tf in self.tempfuncs])
        cnsts = '\n\t\t,'.join(["'%s':%s"%(k,v.src()) for k,v in self.constants.items()])
        return "Temp(constants = {%s}\n\t\t,tempfuncs=[%s])"%(cnsts,tfs)

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    @staticmethod
    def order_tempfuncs(tfs:List['TempFunc'])->List['TempFunc']:
        """
        Prior to storing the list of tempfuncs as a field of the Template,
        order them such that any TF that calls some other TF occurs afterwards
        """

        G = nx.DiGraph() # each template has a DAG, with tempfuncs as nodes
        G.add_nodes_from([tf.name for tf in tfs])

        tf_dict = {tf.name : tf for tf in tfs}

        for tf in tfs:
            for tfa in tf.args:
                if tfa.name in tf_dict.keys():
                    G.add_edge(tfa.name,tf.name)
        return topsort_with_dict(G,tf_dict)

    def apply_template(self
                      ,query_dic  : Dict[str,Any]
                      ) -> Dict[str,Any]:
        """
        This should really be done by creating a computational graph...(DART?)
        """
        consts = {k:v.val for k,v in self.constants.items()}
        dic = {**consts,**query_dic} # MERGE

        for tf in self.tempfuncs:
            dic[tf.name] = tf.apply_func(dic) #type: ignore

        return dic

    def add_temp(self
                ,sqlite_pth   : str
                ,rule_id : int
                ) -> None:
        """
        Add to sqlite meta.db
        """
        for i,tf in enumerate(self.tempfuncs):
            tf.add_tempfunc(sqlite_pth,rule_id,i)
        for j,(constname,const) in enumerate(self.constants.items()):
            const.add_const(sqlite_pth,rule_id,j,constname)

#############################################################################
class TempFunc(object):
    """
    - This object represents a particular function call.
    - Should be given a unique name
    - Each arg refers to the namespace (option of indexing a tuple)
    """
    def __init__(self
                ,func     : Callable
                ,name     : Optional[str] = None
                ,args     : List[Arg]     = []
                ) -> None:
        assert callable(func), 'Not a func: '+str(func)
        self.func     = Func(func)

        if name is None:
            name = self.func.name

        self.name     = name
        self.args     = args

    def src(self)->str:
        args = '\n\t\t,'.join(a.src() for a in self.args)
        fmtargs = [self.name,self.name,args]
        return "TempFunc(funcname='{0}'\n\t\t\t,name='{1}'\n\t\t\t,args=[{2}])".format(*fmtargs)

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def apply_func(self
                  ,curr_dict : Dict[str,Any]
                  ) -> Dict[str,Any]:
        """
        Take a TempFunc's function and wrap it such that it can accept a namespace
            dictionary. The information about how the function interacts with the
            namespace is contained in the TempFunc's args.
        """
        inputvars = [arg.arg_get(curr_dict) for arg in self.args]

        try:
            return self.func.apply(*inputvars)
        except Exception as e:
            msg = '\tApplying func %s in tempfunc %s:\n\t'%(self.func.name,self.name)
            raise ExternalError(msg + traceback.format_exc())

    def add_tempfunc(self
                    ,sqlite_pth  : str
                    ,trans_id    : int
                    ,tf_id       : int
                    ) -> None:
        """
        Add to meta.db
        """
        func_id = func_to_ind(sqlite_pth,self.name) # type: ignore
        cols    = ['rule_id','tempfunc_id','func_id','name']
        q       = mkInsCmd('tempfunc',cols,sqlite=True)
        sqlite_execute(sqlite_pth,q,[trans_id,tf_id,func_id,self.name])
        for i,a in enumerate(self.args):
            a.add_arg(sqlite_pth,'temp',trans_id,tf_id,i)

    @staticmethod
    def maybe_index(d   : dict
                   ,arg : Tuple[str,Optional[int]]
                   ) -> Any:
        """
        If TempFunc arg specifies an index, then we are expecting a tuple and index it
        """
        if arg[1] is None:
            return d[arg[0]]
        else:
            return d[arg[0]][arg[1]]

class Const(object):
    """
    Embed a constant value of any type into the namespace for a given rule.
    """
    def __init__(self,val:Any)->None:
        self.val = val

    def src(self)->str:
        return 'Const(%s)'%repr(self.val)

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def get_const(self,*args: Any)->Any:
        return self.val

    def get_datatype(self)->str:
        return str(self.val.__class__)

    def add_const(self
                 ,sqlite_pth : str
                 ,trans_id   : int
                 ,const_id   : int
                 ,const_name : str
                 )->None:
        """
        Add to meta.db
        """
        datatype = self.get_datatype()
        cols     = ['rule_id','const_id','name','datatype','val']
        q        = mkInsCmd('constants',cols,sqlite=True)
        sqlite_execute(sqlite_pth,q,[trans_id,const_id,const_name,datatype,str(self.val)])

################################################################################################
################################################################################################

#########################
# Special Temp shortcuts
#-----------------------
def identity(x:Any)->Any:
    return x

def SimpleTemp(func    : Callable
              ,inputs  : List[str] = []
              ,outputs : List[str] = []
              ,const   : Dict[str,Const] = {}
              ) -> Temp:
    """
    For a rule that has one function in its template
    """
    main_func = TempFunc(func,'main',args=[Arg(x) for x in inputs])
    if len(outputs)==1:
        exported = [TempFunc(identity,outputs[0],[Arg('main')])] # don't index if output is not expected to be a tuple
    else:
        exported  = [TempFunc(identity,x,[Arg('main',i)])
                        for i,x in enumerate(outputs)]
    return Temp(tempfuncs = [main_func]+exported
               ,constants = const)

def SimplePipe(funcs   : List[Callable]
              ,inputs  : List[str]
              ,outputs : List[str]
              ,const   : Dict[str,Const] = {}
              ) -> Temp:
    """
    A series of functions where the output of each function is fed to the
    following function in the pipeline. You must specify the names of the
    inputs and outputs to the whole process
    """
    init_func = TempFunc(funcs[0],args=[Arg(x) for x in inputs])
    remaining = [TempFunc(funcs[i],args=[Arg(funcs[i-1].__name__)]) for i in range(1,len(funcs))]
    if len(outputs)==1:
        exported = [TempFunc(identity,outputs[0],[Arg(funcs[-1].__name__)])] # don't index if output is not expected to be a tuple
    else:
        exported  = [TempFunc(identity,x,[Arg(funcs[-1].__name__,i)])
                        for i,x in enumerate(outputs)]
    return Temp(tempfuncs = [init_func] + remaining + exported
               ,constants = const)



from typing   import List,Tuple
from ase.data import chemical_symbols # type: ignore
from imp      import load_source
from re       import findall
from os       import environ
from os.path  import join

################################################################################

def parse_keld() -> Tuple[List[str]
                         ,List[int]
                         ,List[str]
                         ,List[str]
                         ,List[int]
                         ,List[int]
                         ,List[float]
                         ,List[str]
                         ,List[float]]:
    """
    Extracts information of materials into triples for struct_dataset_element
    If the species doesn't exist in the database, we need to add it.
    """

    rootpath  = environ['DBGEN_ROOT']
    keld      = load_source('keld',join(rootpath,'data/data_solids_wPBE.py'))
    keld_dict = keld.data                   # type: ignore
    cols      = ['lattice parameter'
                ,'cohesive energy'
                ,'bulk modulus'
                ,'magmom']

    struct_dict = {'fcc'    :'A_1_a_225'
                  ,'bcc'    :'A_1_a_229'
                  ,'hcp'    :'A_2_c_194'
                  ,'diamond':'A_2_a_227'
                  ,'b1'     :'AB_1_a_b_225'  # rocksalt
                  ,'b2'     :'AB_1_a_b_221'  # cesium chloride
                  ,'b3'     :'AB_1_a_c_216'} # zincblende


    dataset     = [] # type: List[Tuple[str,str,float]] # speciesname, property, value
    bulkspecies = [] # type: List[Tuple[str,str]]       # speciesname, pure_struct
    composition = [] # type: List[Tuple[str,int,int]]   # speciesname,number,count

    for k,v in keld_dict.items():
        elemstr,crystal = k.split('-')                  # 'LiH','b1'
        pure_struct  = struct_dict[crystal]             # 'b1' -> 'AB_1_a_b_225'
        elems = findall('[A-Z][^A-Z]*', elemstr)        # ['Li','H']

        bulkspecies.append((k,pure_struct))
        for elem in elems:
            num = chemical_symbols.index(elem)
            composition.append((k,num,1))

        for c in cols:
            val = v.get(c)
            if val is None:
                val = 0
            dataset.append((k,c,val))                   # ('LiH-b1','magmom',0)

    sp_short,struc     = zip(*bulkspecies)
    sp_mid,numb,count  = zip(*composition)
    sp_long,prop,value = zip(*dataset)
    spacegroups        = [int(x.split('_')[-1]) for x in struc]
    output             = [sp_short, struc,  spacegroups
                         ,sp_mid,   numb,   count
                         ,sp_long,  prop,   value]
    return tuple(map(list,output))# type: ignore

if __name__=='__main__':
    parse_keld()



from typing import Tuple,List,Dict
from networkx import MultiGraph                      # type: ignore
from networkx.algorithms import connected_components # type: ignore
from ase.data import chemical_symbols as symbs       # type: ignore
from pprint import pformat
def identify_ads(G:MultiGraph)->Tuple[List[int],List[int],List[str]
                                     ,List[int],List[str],List[str]]:
    """
    Returns [ATOM_INDEX] [ADSORBATE_ID] [NAME] triples (all lists of length M),
     followed by [ADSORBATE_ID] [NAME] [SITE] triples (all lists of length N < M)

    Atoms are considered to be adsorbates if they are in the set {H,O,C,N}
    These atoms are grouped by bonding and adsorbates are assigned arbitrary IDs
    THAT START AT 1 (Adsorbate ID = 0 means not an adsorbate in the atom table)
    """
    # Initialize Variables
    #--------------------
    ads_nodes = [] # type: List[int]
    adsnames  = []
    adssites  = []
    atomdata  = []
    m_bonds   = {} # type: Dict[int,int]
    ads_ind   = 1
    nonmetals = [1,4,6,7,8,9]
    has_metal = False # if we don't see any metal atoms, then there are no adsorbates (e.g. pure water structure)
                      # need to think about graphite / graphene cases
    # Constants
    #----------
    name_dict = {(1,):'H'
                ,(8,):'O'
                ,(7,):'N'
                ,(6,):'C'
                ,(8,8):'O2'
                ,(7,7):'N2'
                ,(6,6):'C2'
                ,(1,1):'H2'
                ,(1,7):'NH'
                ,(1,8):'OH'
                ,(1,1,7):'NH2'
                ,(1,1,1,7):'NH3'
                ,(1,7,7):'NNH'
                ,(1,1,7,7):'NNH2'
                ,(1,1,1,7,7):'NNH3'
                ,(1,1,8):'H2O'
                ,(6,8):'CO'
                ,(1,6,8):'CHO'
                ,(1,8,8):'OOH'
                ,(6,6,8,8):'OCCO'
                ,(1,6,6,8,8):'OCCHO'
                }
    site_dict = {1:'top'
                ,2:'bridge'
                ,3:'threefold'
                ,4:'fourfold'
                }
    # Create subgraph with only nonmetals (with 3 or fewer metal bonds)
    for i,data in G.nodes(data=True):
        if data['number'] in nonmetals:
            metal_bonds = 0
            for n in G.neighbors(i):
                if G.nodes[n]['number'] not in nonmetals:
                    metal_bonds+=1
            m_bonds[i] = metal_bonds
            if metal_bonds < 5:
                ads_nodes.append(i)
        else:
            has_metal = True
    if not has_metal:
        return [],[],[],[],[],[]
    else:

        ads = connected_components(G.subgraph(ads_nodes))
        for a in ads:
            elems = [G.nodes[n]['number'] for n in a]
            name  = name_dict.get(tuple(sorted(elems)))
            if name is not None:
                m_tot = sum(map(m_bonds.get,a)) # type: int
                site  = site_dict.get(m_tot,'other')
                adsnames.append(name)
                adssites.append(site)
                atomdata.extend([(G.nodes[n]['index'],ads_ind,name) for n in a])
                ads_ind+=1
            else:
                pass
                #print('could not find an adsorbate corresponding to '+pformat(elems))
            #summing number of metal bonds those atoms collectively have
        #import pdb;pdb.set_trace()
        if not atomdata:
            return [],[],[],[],[],[]
        else:
            inds,atom_ids,atom_names = tuple(map(list,zip(*atomdata)))
            ads_inds = list(range(1,len(adsnames)+1))
            return inds,atom_ids,atom_names,ads_inds,adsnames,adssites



from typing import Dict
from dbgen.core.parsing import parse_line

def parse_psp_qe(log :str) -> str:
    """
    Returns simplified string representation of the pseudopotnetial used in a
    DFT calculation
    """
    pseudoline = parse_line(log,'pseudo dir',0)
    if pseudoline is None:
        raise ValueError('malformed QE log file? can''t find "pseudo dir"')
    else:
        return pseudoline.split(":")[1].strip()



from typing import List
from ase import Atoms                      # type: ignore
from ase.data import covalent_radii,vdw_radii        # type: ignore
import numpy as np                         # type: ignore
from ase.neighborlist import NeighborList  # type: ignore
from json import dumps
from math import exp,isnan

def geometric(atoms:Atoms)->str:
    """
    If two atoms have a distance less than their covalent_radii sum, they're bonded
    (B.O. = e^(1-distance))
    """
    output     = [] # type: List


    pos        = atoms.get_positions()
    cell       = atoms.get_cell()
    cutoffs    = [covalent_radii[x]+0.25 for x in atoms.get_atomic_numbers()]
    nl = NeighborList(cutoffs           = cutoffs
                     ,skin              = 0
                     ,self_interaction  = False
                     ,bothways          = True)
    atoms.set_pbc([1,1,1])
    nl.update(atoms)
    for i in range(len(atoms)-1): # never need to check last one
        for j,(x,y,z) in zip(*nl.get_neighbors(i)):
            if i <= j:
                d = np.linalg.norm(pos[i]-pos[j]-np.dot([x,y,z],cell))
                pseudo_bo = round(3*exp(1-d),2)
                output.append([int(i),int(j),pseudo_bo,int(x),int(y),int(z)])

    return dumps(output)

if __name__=='__main__':
    import sys
    from ase.io import read # type: ignore
    print(geometric(read(sys.argv[1])))



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_diag_vasp(log:str) -> Optional[str]:
    """
    Uh oh, guess we don't yet know how to extract which Diagonalization algo was used in VASP
    """
    return None



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_mixing_gpaw(log:str) -> Optional[float]:
    """
    Helpful docstring
    """
    parsed = parse_line(log,'Linear mixing parameter',0)
    if parsed is None:
        raise ValueError
    else:
        raw    = parsed.split(':')[1]
        return round(float(raw),3)



import json
import numpy as np                      # type: ignore
import ase                              # type: ignore
from ase.constraints import FixAtoms    # type: ignore

def json_to_traj(raw_json:str)->ase.Atoms:
    """
    Inverse of traj_to_json
    """

    raw_atoms = json.loads(raw_json)

    atom_data = raw_atoms['atomdata']

    pos = np.array([[a[q] for a in atom_data] for q in ['x','y','z']]).T

    fix = FixAtoms([a['index'] for a in atom_data if a['constrained']])

    atoms = ase.Atoms(numbers     = [a['number'] for a in atom_data]
                     ,cell        = raw_atoms['cell']
                     ,positions   = pos
                     ,magmoms     = [a['magmom'] for a in atom_data]
                     ,tags        = [a['tag'] for a in atom_data]
                     ,constraint  = fix)
    return atoms



import numpy as np  # type: ignore
import ase          # type: ignore

def get_vacuum(atoms : ase.Atoms) -> float:
    """
    Same method as the one used in get_system_type.
    Assumes that z axis is normal to surface (would need PMG functions to
        reorient and make this function more robust)
    """
    c     = np.linalg.norm(atoms.get_cell()[2])
    delta = 0.0
    zs    = sorted([a.z for a in atoms])
    z     = zs[0]

    for z_next in zs:
        delta = max(delta, z_next - z)
        z = z_next

    return c - delta



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_mixing_qe(log:str) -> Optional[float]:
    """
    Get mixing beta from QE log file
    """

    parsed = parse_line(log,'mixing_beta',0) or parse_line(log,'mixing beta',0)  # different QE versions?
    if parsed is None:
        raise ValueError
    else:
        raw    = parsed.split("=")[1]
        return round(float(raw),3)



from typing import Dict,Tuple
import ast

from dbgen.core.parsing import parse_line

def get_kpts_gpaw(trip:Tuple[str,str,str]) -> Tuple[int,int,int]:
    """
    docstring
    """
    log,_,_    = trip
    parsed = parse_line(log,'k-points: ')
    if parsed is None:
        return (1,1,1) # default
    else:
        raw    = parsed.split(': ')[1].split()
        return (int(raw[0]),int(raw[2]),int(raw[4]))



# External Modules
from typing      import List,Optional,Dict,FrozenSet,Set,Tuple
from pprint      import pformat
from copy        import deepcopy
from abc         import abstractmethod
from collections import defaultdict,OrderedDict
from networkx    import DiGraph                                 # type: ignore
from networkx.algorithms.simple_paths import all_simple_paths   # type: ignore
from networkx.algorithms.dag          import descendants        # type: ignore
from networkx.algorithms              import simple_cycles      # type: ignore

# Internal modules
from dbgen.core.lists                     import flatten
from dbgen.support.datatypes.attr         import Attr
from dbgen.support.datatypes.table        import Table,FK as OldFK
from dbgen.support.datatypes.constraint   import EQ,NE,LT,GT,Constraint
from dbgen.support.datatypes.insertupdate import Insert,Update

"""
Defines the FK, Object, Join, From, and Universe classes
"""

class Object(object):
    """
    parent      - What is this object defined by? If nothing, it gets its own
                  independent ID, otherwise it requires IDs of parents to be constructed
    attr        - Simple attributes of the object (int/text/float)
    uniqattr    - List of attr names which should have UNIQUE constraint applied
    many_to_one - indicates a many-one relationship with parents

    Current Problems
        - cannot put a not null constraint on component FK
            (specify 'necessary component')
    """
    def __init__(self
                ,name        : str
                ,desc        : str
                ,parents     : List[str]  = []
                ,components  : List[str]  = []
                ,attr        : List[Attr] = []
                ,uniqattr    : List[str]  = []
                ,many_to_one : bool       = False
                ) -> None :

        self._name        = name
        self._desc        = desc
        self._parents     = parents
        self._components  = components
        self._many_to_one = many_to_one

        # Handle attributes
        #------------------
        attrs = []
        for a in attr:
            assert a.name[0] != '_' # no collisions with internal fields
            a.obj = self._name
            if a.name in uniqattr:
                a.uniq = True
            attrs.append(a)
            setattr(self,a.name,a)

        self._attr = attrs

        # Integrity checks
        #-----------------
        if many_to_one:
            assert len(parents) > 0

    # Basic Methods
    #--------------
    def __str__(self) -> str:
        return 'Object(%s)'%(self._name)

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self,other : object)->bool:
        if isinstance(other,Object):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __hash__(self)->int:
        return hash(self._name)
    #
    # # Other Methods
    # #--------------
    # def get(self,attr : str)->Optional[Attr]:
    #     """
    #     Safe access of attributes
    #     """
    #     for a in self._attr:
    #         if a.name == attr:
    #             return a
    #     return None
    #
    # def get_unsafe(self,attr:str)->Attr:
    #     x = self.get(attr)
    #
    #     if x is None:
    #         msg = "Called getUnsafe when Attr %s not present in Object %s"
    #         raise ValueError(msg%(attr,self._name))
    #     else:
    #         return x
    #
    # def select(self
    #           ,schema : 'Schema'
    #           ,attrs  : Optional[List[str]] = None
    #           ) -> List[Attr]:
    #     if attrs is None:
    #         return self._pks(schema)
    #     else:
    #         return list(filter(None,map(self.get,attrs)))
    #
    # def _pks(self,schema:'Schema')->List[Attr]:
    #     """
    #     Return a list of Attr that are PKs as well as any related FK objects
    #
    #     Keep the data associated with parent PKs excpet:
    #         - change the owner
    #         - prefix the name with the table it came from
    #         - remove the autoincrement
    #     """
    #     # Create key for this object (not needed if completely dependent)
    #     if self._many_to_one:
    #         ks = [Attr(name=self._name+'_id',pk=True,auto=True,obj=self._alias)]
    #     else:
    #         ks = []
    #
    #     return ks + self._pkfks(schema)
    #
    # def _pkfks(self,schema:'Schema')->List[Attr]:
    #     pkfks = []
    #     for p in schema.get(self._parents):
    #         for k in p._pks(schema):
    #             k.auto = False
    #             k.name = k.obj+'_'+k.name
    #             k.obj  = self._alias
    #             pkfks.append(k)
    #     return pkfks
    #
    # def _fks(self,schema:'Schema')->List['FK']:
    #     FKs = []
    #     for comp in schema.get(self._components):
    #         colnames = [comp._name+'_'+c.name for c in comp._pks(schema)]
    #         _fks     = []
    #         for cn in colnames:
    #             fk = Attr(cn)
    #             fk.obj = self._alias
    #             _fks.append(fk)
    #             setattr(self,cn,fk)
    #         FKs.append(FK(self,_fks,comp,comp._pks(schema)))
    #     return FKs
    #
    # def _fk_attrs(self,schema:'Schema')->List[Attr]:
    #     """Get the from_attr from all of an Object's FKs (so that we can create
    #     the Columns)"""
    #     return flatten([fk.from_attr for fk in self._fks(schema)])
    #
    # def As(self,alias:str)->'Object':
    #     """
    #     Create identical object but with different alias
    #     """
    #     new = deepcopy(self)
    #     new._alias +='_'+alias
    #     return new
    #
    # def _mkTable(self,schema:'Schema')->Table:
    #     cols = flatten([list(x) for x in [self._attr
    #                                      ,self._pks(schema)
    #                                      ,self._fk_attrs(schema)]])
    #     return Table(name = self._name
    #                 ,desc = self._desc
    #                 ,cols = [c._mkCol() for c in cols]
    #                 ,fks  = [fk.to_old_style() for fk in self._fks(schema)])
    #
    # def insert(self
    #           ,schema    : 'Schema'
    #           ,attr_list : Optional[List[str]] = None
    #           ) -> Insert:
    #     if attr_list is None:
    #         return Insert(schema,self._pks(schema))
    #     else:
    #         try:
    #             return Insert(schema,[self.get_unsafe(x) for x in attr_list])
    #         except ValueError as e:
    #             print(e)
    #             import pdb;pdb.set_trace()
    #             raise ValueError
    #
    # def update(self,schema:'Schema',attr_list:List[str])->Update:
    #     return Update(schema,[self.get_unsafe(x) for x in attr_list])
    #
    # def update_component(self
    #                     ,schema : 'Schema'
    #                     ,o      : 'Object'
    #                     ) -> Update:
    #     fks = [fk for fk in self._fks(schema) if fk._to == o]
    #     assert len(fks)==1
    #     return Update(schema,fks[0].from_attr)


################################################################################
class FK(object):
    """
    A Foreign Key in the language of Objects/Attributes is a pair of lists of
    Attributes, each having equal length but referring to different Objects
    """
    def __init__(self
                ,from_attr : List['Attr']
                ,_to       : '_Object'
                ,to_attr   : List['Attr']
                ) -> None:
        # Validate
        #---------
        err1 = 'FK args not of equal length: %s %s'
        assert len(from_attr) == len(to_attr), err1%(from_attr,to_attr)
        err2 = 'FK args share the same object: %s'
        assert from_attr[0].obj != to_attr[0].obj, err2%from_attr[0].obj

        # Store Fields
        #-------------

        self.from_attr = from_attr
        self._to       = _to
        self.to_attr   = to_attr

    def flip(self,_from:'_Object') -> None:
        temp_attr       = self.from_attr
        self.from_attr  = self.to_attr
        self.to_attr    = temp_attr
        self._to        = _from

    def zipped(self) -> str:
        z = zip(self.from_attr,self.to_attr)
        return ' AND '.join(['%s = %s'%(f,t) for f,t in z])

    def to_old_style(self) -> OldFK:
        if self.to_attr[0].obj is None:
            raise ValueError
        else:
            return OldFK([a.name for a in self.from_attr]
                         ,self._to.name
                         ,[a.name for a in self.to_attr])

################################################################################

class _Object(object):
    def __init__(self
                ,name        : str
                ,desc        : str
                ,parents     : List[_Object]  = []
                ,components  : List[_Object]  = []
                ,attr        : List[Attr]     = []
                ,pks         : List[Attr]     = []
                ,fks         : List[FK]       = []
                ,fks_to      : List[FK]       = []
                ,many_to_one : bool           = False
                ) -> None :
        self.name        = name
        self.desc        = desc
        self.parents     = parents
        self.components  = components
        self.attr        = attr
        self.pks         = pks
        self.fks         = fks
        self.fks_to      = fks_to
        self.many_to_one = many_to_one

        # Validate
        #---------
        assert set(pks).issubset(attr)

class Universe(object):
    """
    Unordered, immutable collection of Objects
    """
    def __init__(self,objects:List[Object])->None:
        self.objects   = frozenset(objects)
        self.obj_dict  = {o._name : o for o in objects}
        self.pk_dict   = {} # type: Dict[str,List[Attr]]
        self._obj_dict = {} # type: Dict[str,_Object]
        self._obj_dict = {n:self.upgrade(n) for n in [o._name for o in objects]}
        self._objects  = frozenset(self._obj_dict.values())
        self.graph     = self.mk_graph()
        import pdb;pdb.set_trace()

    def get_pks(self,obj_str:str)->List[Attr]:
        if obj_str in self.pk_dict:
            return self.pk_dict[obj_str]
        else:
            obj = self.obj_dict[obj_str]
            a0  = Attr(name=obj_str+'_id',pk=True,auto=True,obj=obj._name)
            pks = [a0] if (obj._many_to_one or (not obj._parents)) else []
            for p in obj._parents:
                new_pks = [Attr(p+'_'+k.name,obj=obj._name)
                            for k in self.get_pks(p)]
                pks.extend(new_pks)
            return pks

    def upgrade(self,o_str:str)->_Object:
        """
        In light of a fixed Universe, compute richer properties of each Object
        (recursively with memoization)
        """
        if o_str in self._obj_dict: return self._obj_dict[o_str]

        # Get primary keys (recursively, with memoization)
        pks = self.get_pks(o_str)

        o = self.obj_dict[o_str]

        # Get Foreign Keys from this object to its components
        fks = []
        for comp in o._components + o._parents:
            colnames = [comp + '_' + c.name for c in self.get_pks(comp)]
            _fks     = [Attr(cn,obj=o_str) for cn in colnames]
            fks.append(FK(_fks,self.upgrade(comp),self.get_pks(comp)))

        # Get the columns associated with foreign keys from this object
        fk_attrs = flatten([fk.from_attr for fk in fks])

        # Assemble all the information
        _object = _Object(name     = o_str
                      ,desc        = o._desc
                      ,parents     = [self.upgrade(x) for x in o._parents]
                      ,components  = [self.upgrade(x) for x in o._components]
                      ,attr        = self.get_pks(o_str) + fk_attrs + o._attr
                      ,pks         = pks
                      ,fks         = fks
                      ,many_to_one = o._many_to_one)
        self._obj_dict[o_str] = _object
        return _object

    def mk_graph(self)->DiGraph:
        """
        Given parent/component relationships between objects, create a Directed
        Graph that captures the possible flows of information.
        """
        G = DiGraph()
        G.add_nodes_from([o._name for o in self.objects])
        for o1 in self.objects:
            if not o1._many_to_one:
                for o2 in o1._parents:
                    G.add_edge(o2,o1._name)
            for o2 in o1._parents + o1._components:
                    G.add_edge(o1._name,o2)
        return G

    def path(self,o:str,o2:str)->List[List[str]]:
        """
        All possible paths between two objects via FK propagation
        """
        paths = all_simple_paths(self.mk_graph(),o,o2)
        return sorted(list(paths),key=len)

    def get_fk(self,o1:_Object,o2:_Object)->FK:
        """
        Gets a FK between two Objects -- error if more/fewer are found
        """
        fks = [fk for fk in o1.fks if fk._to is o2]  + [fk for fk in o2.fks if fk._to is o1]
        if len(fks) > 1:
            print('Too many fks between %s and %s: %s'%(o1,o2,fks))
            import pdb;pdb.set_trace()
            raise ValueError
        elif len(fks) == 0:
            print('No many fks between %s and %s'%(o1,o2))
            import pdb;pdb.set_trace()
            raise ValueError
        else:
            fk = fks[0]
            if fk._to != o2:
                fk.flip(o1)
            return fk

    def get(self,obj_str:List[str])->List[_Object]:
        return [self.get_(x) for x in obj_str]

    def get_(self,obj_str:str)->_Object:
        return self._obj_dict[obj_str]

    def select(self
              ,attrs       : List[Attr]
              ,constraints : List[Constraint] = []
              ,maybeNull   : List[Object]     = []
              ) -> str:
        allobj    = self.get(sorted(set([a.obj for a in attrs])))
        pk_attrs  = flatten([o.pks for o in allobj])

        # Make from
        cols   = '\n\t,'.join([str(a) for a in attrs])
        consts = '\n\tAND '.join([str(c) for c in constraints])
        c_attrs = flatten([list(c.attrs()) for c in constraints])

        all_attr = attrs + c_attrs
        _from = self.mkFrom(all_attr)

        for j in _from.joins:
            if j.fk._to in maybeNull:
                j.kind = "LEFT"

        return 'SELECT %s\n%s\nWHERE %s'%(cols,_from,consts or '1')

    def mkFrom(self,attrs:List[Attr])->From:
        """
        Construct a FROM clause that joins the right tables in order to get
        access to a set of attributes mentioned by some query
        """
        objects  = self.get([a.obj for a in attrs])

        found    = False
        for obj in objects:
            desc = descendants(self.graph,obj.name)
            if all([o.name in desc for o in objects if o != obj]):
                found = True
                first = obj

        if not found:
            raise ValueError('You screwed up')

        joins = []
        for o in objects:
            if o != first:
                paths = self.path(first.name,o.name)
                if len(paths)>1:
                    print("Warning, more than one path found between %s and %s"%(first,o))
                path = paths[0]
                assert path[0] == first.name
                curr_obj = first # initialize with this

                for s in path[1:]: # first element = first._name
                    next_obj = self._obj_dict[s]
                    fk = self.get_fk(curr_obj,next_obj)
                    joins.append(Join(next_obj,fk))
                    curr_obj = next_obj

        return From(first,joins)
#
#
#
class From(object):
    """
    FROM clause
    """
    def __init__(self
                ,obj   : _Object
                ,joins : List[Join] = []
                ) -> None:
        self.obj   = obj
        self.joins = joins

    # Basic Methods
    #--------------
    def __str__(self) -> str:
        raise NotImplementedError
        return 'FROM %s AS %s '%(self.obj._name,self.obj._name) + ''.join(['\n%s'%j for j in self.joins])

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self,other : object)->bool:
        if isinstance(other,Object):
            return self.__dict__ == other.__dict__
        else:
            return False


class Join(object):
    """
    JOIN <A> AS <B> ON <C> = <D> AND ...
    """
    def __init__(self
                ,obj  : _Object
                ,fk   : FK
                ,kind : str     = 'INNER'
                ) -> None:
        self.obj  = obj
        self.fk   = fk
        self.kind = kind
    # Basic Methods
    #--------------
    def __str__(self) -> str:
        join   = '%s JOIN'%self.kind
        to_obj = self.fk._to
        if to_obj is None:
            raise ValueError
        else:
            tab    = '%s AS %s'%(to_obj.name,to_obj.name)
            return '\n\t%s %s \n\t\tON %s'%(join,tab,self.fk.zipped())



import typing

class DataType(object):
    def toJSON(self)->dict: raise NotImplementedError()

class AnyType(DataType):
    def toJSON(self)->dict:
        return {'tag':'AnyType'}

    def __str__(self)->str:
        return 'Any'

    def __repr__(self)->str:
        return self.__str__()

class NoneType(DataType):
    def toJSON(self)->dict:
        return {'tag':'NoneType'}

    def __str__(self)->str:
        return 'None'

    def __repr__(self)->str: return self.__str__()

class Base(DataType):
    def __init__(self,unBase:str)->None:
        self.unBase = unBase

    def toJSON(self)->dict:
        return {'tag':'Base','unBase':self.unBase}

    def __str__(self)->str:
        return '"%s"'%self.unBase

    def __repr__(self)->str: return self.__str__()

class TypeVar(DataType):
    def __init__(self,name:str)->None:
        self.name = name

    def toJSON(self)->dict:
        return {'tag':'TypeVar','name':self.name}

    def __str__(self)->str:
        return '"%s"'%self.name


class Callable(DataType):
    def __init__(self
                ,args:typing.List[DataType]
                ,out:DataType
                )->None:
        self.c_args= args
        self.out = out

    def toJSON(self)->dict:
        return {'tag':'Callable'
                ,'cArgs':[x.toJSON() for x in self.c_args]
               ,'out':self.out.toJSON()
               }

    def __str__(self)->str:
        ar = ' -> '
        return ar.join(map(str,self.c_args))+ar+str(self.out)

    def __repr__(self)->str:
        return self.__str__()

class Union(DataType):
    def __init__(self,args:typing.List[DataType])->None:
        self.args=args

    def toJSON(self)->dict:
        return {'tag':'Union'
               ,'uArgs':[x.toJSON() for x in self.args]}

    def __str__(self)->str:
        return '{%s}'%(','.join(sorted(map(str,self.args))))

    def __repr__(self)->str:
        return self.__str__()

class Tuple(DataType):
    def __init__(self,args:typing.List[DataType])->None:
        self.args= args
    def toJSON(self)->dict:
        return {'tag':'Tuple'
               ,'tArgs':[x.toJSON() for x in self.args]}


    def __str__(self)->str:
        return '(%s)'%(','.join(map(str,self.args)))

    def __repr__(self)->str: return self.__str__()

class List(DataType):
    def __init__(self,content:DataType)->None:
        self.content=content

    def toJSON(self)->dict:
        return {'tag':'List'
               ,'unList':self.content.toJSON()}

    def __str__(self)->str:
        return '[%s]'%(str(self.content))

    def __repr__(self)->str:
        return self.__str__()

class Dict(DataType):
    def __init__(self,fromT:DataType,toT:DataType)->None:
        self.key = fromT
        self.val = toT

    def toJSON(self)->dict:
        return {'tag':'Dict'
                ,'key': self.key.toJSON()
                ,'val': self.val.toJSON()}

    def __str__(self)->str:
        return '{ %s : %s }'%(str(self.key),str(self.val))

    def __repr__(self)->str: return self.__str__()



from typing import List,Tuple

import json
from ase import Atoms # type: ignore

def get_atom(atomsjson : str
            )->Tuple[List[str],List[int],List[int]
                        ,List[float],List[float],List[float]
                        ,List[int],List[float],List[int]]:
    """
    Helpful
    """
    atomdata = json.loads(atomsjson)['atomdata']
    output   = [(atomsjson
                ,a['index']
                ,a['number']
                ,a['x']
                ,a['y']
                ,a['z']
                ,a['constrained']
                ,a['magmom']
                ,a['tag']
                ) for a in atomdata]


    return tuple(map(list,zip(*output))) # type: ignore



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_nmix_vasp(outcar:str) -> Optional[float]:
    """
    What vasp keyword corresponds to this, again?
    """
    return None



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_spinpol_vasp(outcar:str) -> Optional[float]:
    """
    Docstring
    """

    parsed = parse_line(outcar,'ISPIN',0)
    if parsed is None:
        raise ValueError
    else:
        return int('1' in parsed)



from typing import Any,Dict

def get_(d       : Dict[Any,Any]
        ,key     : str
        ,default : Any              = None
        ) -> Any:
    """
    Gets a key from a dictionary
    """
    return d.get(key,default)



from typing import Dict
from dbgen.core.parsing import parse_line

def get_dipole_qe(log:str) -> int:
    """
    Parse pw.inp file to see if dipole correction was applied
    """
    return int('Computed dipole' in log)



from typing import Dict,List,Tuple,Optional
# External modules
MbFloat = Optional[float]
ForceTuple = Tuple[int,int,MbFloat,MbFloat,MbFloat]
from ase        import Atoms # type: ignore
from string     import digits
from os.path    import join,exists
from os         import listdir
# Internal modules
from dbgen.core.numeric import roundfloat
################################################################################
def get_traj_gpaw_lazy(a_l_p : Tuple[str,str,str]) -> List[Tuple[Atoms,int,float,List[ForceTuple],int]]:
    """
    we don't do optimizations in gpaw???

    Returns list of triples to represent the trajectory of a structure during
    an optimization. The ASE structure, the step number along the trajectory,
    and the energy. These are parsed from a GPAW log file.
    """
    _,log,_ = a_l_p
    # Constants
    #---------
    output   = []
    forces   = [] # type: List[ForceTuple]
    stepnum  = 0
    FORCE = False
    CELL  = False
    POS   = False

    # Get log file
    #-------------
    lines = log.split('\n')
    for line in lines: # read NOT lazily
        # MAIN LOOP
        #----------
        if 'Positions:' in line:
            POS = True
            atomdata=[]
        elif 'Number of atoms' in line:
            num_atoms = int(line.split()[-1])
        elif POS:
            ind,sym,x,y,z = line.split()
            atomdata.append((sym,(roundfloat(x),roundfloat(y),roundfloat(z))))
            if int(ind)+1 == num_atoms:
                POS = False
        elif 'periodic' in line:
            CELL = True
            cellvecs =[]
        elif CELL:
            cellvecs.append([roundfloat(x) for x in line.split()[3:6]])
            if '3. axis' in line:
                CELL = False
                symbs = [a[0] for a in atomdata]
                pos   = [a[1] for a in atomdata]
                atoms = Atoms(symbols=symbs,positions=pos,cell=cellvecs)
                atoms.wrap()
        elif 'Forces in eV/Ang' in line:
            FORCE = True
            forces = []
        elif FORCE:
            if line.strip()=='':
                FORCE = False
            else:
                try:
                    i,_,fx,fy,fz = line.split()
                    forces.append((stepnum,int(i),roundfloat(fx),roundfloat(fy),roundfloat(fz)))
                except ValueError:
                    FORCE = False
                    forces =[]
        elif 'Extrapolated' in line:
            eng =float(line.split()[-1])

    if not forces:
        forces = [(0,i,None,None,None) for i in range(num_atoms)]

    output.append((atoms,stepnum,eng,forces,1))
    return output
